/*
 * Complex number arithmetics.
 *
 * Grisha Sitkarev, <sitkarev@unixkomi.ru>
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "complex.h"

complex_t c_num(double re, double im)
{
	complex_t res;

	res.re = re;
	res.im = im;

	return res;
}

complex_t c_abs(complex_t a)
{
	complex_t res;

	res.re = fabs(a.re);
	res.im = fabs(a.im);

	return res;
}

complex_t c_div(complex_t a, complex_t b)
{
	complex_t res;
	double real, imag, div;

	real = (a.re * b.re) + (a.im * b.im);
	imag = (a.im * b.re) - (a.re * b.im);
	div = (b.re * b.re) + (b.im * b.im);
	res.re = real / div;
	res.im = imag / div;

	return res;
}

complex_t c_sum(complex_t a, complex_t b)
{
	complex_t res;

	res.re = a.re + b.re;
	res.im = a.im + b.im;

	return res;
}

complex_t c_sub(complex_t a, complex_t b)
{
	complex_t res;

	res.re = a.re - b.re;
	res.im = a.im - b.im;

	return res;
}

complex_t c_mult(complex_t a, complex_t b)
{
	complex_t res;

	res.re = (a.re * b.re) - (a.im * b.im);
	res.im = (a.im * b.re) + (a.re * b.im);

	return res;
}

complex_t c_scale(complex_t a, double scale)
{
	complex_t res;

	res.re = a.re * scale;
	res.im = a.im * scale;

	return res;
}

double c_magnitude(complex_t a)
{
	return hypot(a.re, a.im);
}

double c_phase(complex_t a)
{
	if ((a.im == 0.0) && (a.re == 0.0))
		return 0.0;

	return atan2(a.im, a.re);
}

complex_t c_polar(double magnitude, double phase)
{
	complex_t res;

	res.re = magnitude * cos(phase);
	res.im = magnitude * sin(phase);

	return res;
}

complex_t c_conjugate(complex_t a)
{
	complex_t res;

	res.re = a.re;
	res.im = -a.im;

	return res;
}

complex_t c_sqrt(complex_t a)
{
	complex_t res;

	res.re = sqrt((sqrt(a.re * a.re + a.im * a.im) + a.re) / 2.0);

	if (a.im < 0.0)
		res.im = -1.0 * sqrt((sqrt(a.re * a.re + a.im * a.im) - a.re) / 2.0);
	else
		res.im = sqrt((sqrt(a.re * a.re + a.im * a.im) - a.re) / 2.0);

	return res;
}

complex_t c_inverse(complex_t a)
{
	double b;
	complex_t res;

	if (a.re == 0.0 && a.im == 0.0) {
		res.re == INFINITY;
		res.im == INFINITY;
	} else {
		b = a.re * a.re + a.im * a.im;
		res.re = a.re / b;
		res.im = a.im / b;
	}

	return res;
}

complex_t c_sin(complex_t a) 
{
	complex_t res;

	if (a.im == 0.0){
    	/* avoid returing negative zero (-0.0) for the imaginary part  */ 
    		res.re = sin(a.re);
    		res.im = 0.0;
 	} else {
		res.re = sin(a.re) * cosh(a.im);
		res.im = cos(a.re) * sinh(a.im);
 	}

	return res;
}

complex_t c_cos(complex_t a)
{
	complex_t res;

	if (a.im == 0.0) {
	/* avoid returing negative zero (-0.0) for the imaginary part  */ 
		res.re = cos(a.re);
		res.im = 0.0;
	} else {
		res.re = cos(a.re) * cosh(a.im);
		res.im = sin(a.re) * sinh(-1.0 * a.im);
	}

	return res;
} 

complex_t c_asin(complex_t a) 
{
	double x, y, y2; 
	double r, s; 
	double A, B; 
	double real, imag;
	double D, Apx, Am1;
	
	const double A_crossover = 1.5;
	const double B_crossover = 0.6417;
	
	complex_t res;
	
	if (a.im == 0.0) {
		res.re = asin(a.re);
		res.im = 0.0;
	} else {
		x = fabs(a.re);
		y = fabs(a.im);
		y2 = y * y;
		r = hypot(x + 1.0, y);
		s = hypot(x - 1.0, y);
		A = 0.5 * (r + s);
		B = x / A;

		if (B <= B_crossover)
			real = asin(B);
		else {
			if (x <= 1.0) {
				D = 0.5 * (A + x) * (y2 / (r + x + 1.0) + (s + (1.0 - x)));
				real = atan(x / sqrt(D));
			} else {
				Apx = A + x;
				D = 0.5 * (Apx / (r + x + 1.0) + Apx / (s + (x - 1.0)));
				real = atan(x / (y * sqrt(D)));
			}
        	}

		if (A <= A_crossover) {
			if (x < 1.0)
				Am1 = 0.5 * (y2 / (r + (x + 1.0)) + y2 / (s + (1.0 - x)));
			else
				Am1 = 0.5 * (y2 / (r + (x + 1.0)) + (s + (x - 1.0)));

			imag = log1p(Am1 + sqrt(Am1 * (A + 1.0)));
		} else
			imag = log(A + sqrt(A * A - 1.0));

		res.re = (a.re >= 0.0) ? real : -real;
		res.im = (a.im >= 0.0) ? imag : -imag;
	}

	return res;
}

complex_t c_acos(complex_t a)
{
	double x, y, y2;
	double r, s;
	double A, B;
	double real, imag;
	double D, Apx, Am1;

	const double A_crossover = 1.5;
	const double B_crossover = 0.6417;

	complex_t res;

	if (a.im == 0.0) {
		res.re = acos(a.re);
		res.im = 0.0;
	} else {
		x = fabs(a.re);
		y = fabs(a.im);
		y2 = y * y;
		r = hypot(x + 1.0, y);
		s = hypot(x - 1.0, y);
		A = 0.5 * (r + s);
		B = x / A;

		if (B <= B_crossover)
			real = acos(B);
		else {
			if (x <= 1.0) {
				D = 0.5 * (A + x) * (y2 / (r + x + 1.0) + (s + (1.0 - x)));
				real = atan(sqrt(D) / x);
			} else {
				Apx = A + x;
				D = 0.5 * (Apx / (r + x + 1.0) + Apx / (s + (x - 1.0)));
				real = atan((y * sqrt(D)) / x);
			}
		}

		if (A <= A_crossover) {
			if (x < 1.0)
				Am1 = 0.5 * (y2 / (r + (x + 1.0)) + y2 / (s + (1.0 - x)));
			else
				Am1 = 0.5 * (y2 / (r + (x + 1.0)) + (s + (x - 1.0)));

			imag = log1p(Am1 + sqrt(Am1 * (A + 1.0)));
		} else
			imag = log(A + sqrt(A * A - 1.0));

		res.re = (a.re >= 0.0) ? real : M_PI - real;
		res.im = (a.im >= 0.0) ? -imag : imag;
	}

	return res;
}
